/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
static char *AtFSid = "$Header: log.c[109.0] Wed Nov 24 03:47:11 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <stdio.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include "wfs.h"
#include "global.h"
#include "util.h"

static void *dlog_callout;

static int idle = 0;
static int closed = 1;
static FILE *logfp;
#ifdef _POSIX_SOURCE
static char linebuf[BUFSIZ];
#endif

static int
dlog_open()
{
	char path[MAXPATHLEN];

	sprintf(path, "%s/log/wwfs.log", cs_topdir);
	logfp = fopen(path, "a+");
	if (logfp == NULL) {
		syslog(LOG_WARNING, "%s: %m", path);
		return -1;
	}
	closed = 0;
#ifdef _POSIX_SOURCE
	setvbuf(logfp, linebuf, _IOLBF, sizeof(linebuf));
#else
	setlinebuf(logfp);
#endif
	return 0;
}

#ifdef __STDC__
void
dlog(char *fmt, ...)
{
	va_list ap;
	char *buf, *timestr;
	time_t t = gettime();
	static time_t last_t = 0;
	static char *last_ctime = 0;

	buf = so_getbuf(0);
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	if (!cs_topdir) {
		/* as a last resort.. */
		syslog(LOG_WARNING, buf);
		return;
	}
	idle = 0;
	if (closed) {
		if (dlog_open() < 0) return;
	}
	if (t != last_t) {
		last_ctime = ctime(&t);
		last_t = t;
	}
	fprintf(logfp, "%15.15s %s csd[%d]: %s\n",
		last_ctime+4, cs_hostname, cs_pid, buf);
}
#else

void
dlog(va_alist)
va_dcl
{
	va_list ap;
	char *fmt, *buf;
	time_t t = gettime();
	static time_t last_t = 0;
	static char *last_ctime = 0;

	buf = so_getbuf(0);
	va_start(ap);
	fmt = va_arg(ap, char *);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	if (!cs_topdir) {
		/* as a last resort.. */
		syslog(LOG_WARNING, buf);
		return;
	}
	idle = 0;
	if (closed) {
		if (dlog_open() < 0) return;
	}
	if (t != last_t) {
		last_ctime = ctime(&t);
		last_t = t;
	}
	fprintf(logfp, "%15.15s %s csd[%d]: %s\n",
		last_ctime+4, cs_hostname, cs_pid, buf);
}
#endif

/*
 * automagically close trace file if no activity has been observed recently.
 */
static void
dlog_gc()
{
	if (!closed) {
		idle += 60;
		if (idle > 120) {
			fclose(logfp);
			closed = 1;
		}
	}
	dlog_callout = timeout_set(gettime()+60, dlog_gc, 0);
}

void
dlog_start()
{
	FILE *fp;
	char path[MAXPATHLEN];

	sprintf(path, "%s/log/csd.pid", cs_topdir);
	fp = fopen(path, "w");
	if (fp) {
		fprintf(fp, "%d", cs_pid);
		fclose(fp);
	} else {
		syslog(LOG_WARNING, "%s: %m", path);
	}

	dlog_callout = timeout_set(gettime()+60, dlog_gc, 0);
}

void
dlog_stop()
{
	if (!closed) {
		fclose(logfp);
		closed = 1;
	}
}

#if 0	/* XXX for production use */
void
errno_diag()
{
	syslog(LOG_WARNING, "%m");
}
#endif

void
errno_log(file, line)
char *file;
int line;
{
	extern char *sys_errlist[];
	dlog("%s:%d, %s", file, line, sys_errlist[errno]);
}

static int ts;

void
talk_start()
{
	sockaddr_in from;
	int ret;

	ts = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	from.sin_family = AF_INET;
	from.sin_addr.s_addr = htonl(INADDR_ANY);
	from.sin_port = htons(0);
	ret = bind(ts, (struct sockaddr *) &from, sizeof (from));
	if (ret < 0) {
		errno_diag();
	}
}

#ifdef __STDC__
void
talk(long ip_addr, char *fmt, ...)
{
	va_list ap;
	char *buf;
	sockaddr_in to;

	buf = so_getbuf(0);
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	to.sin_family = AF_INET;
	to.sin_port = htons(8001);	/* XXX */
	to.sin_addr.s_addr = ip_addr;
	sendto(ts, buf, strlen(buf)+1, 0, (struct sockaddr *)&to, sizeof(to));
}

#else

void
talk(va_alist)
va_dcl
{
	va_list ap;
	char *fmt, *buf;
	sockaddr_in to;
	long ip_addr;

	buf = so_getbuf(0);
	va_start(ap);
	ip_addr = va_arg(ap, long);
	fmt = va_arg(ap, char *);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	to.sin_family = AF_INET;
	to.sin_port = htons(8001);	/* XXX */
	to.sin_addr.s_addr = ip_addr;
	sendto(ts, buf, strlen(buf)+1, 0, (struct sockaddr *)&to, sizeof(to));
}

#endif
