/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* Cluster Server; part of Worldwide File System */
static char *AtFSid = "$Header: main.c[109.0] Wed Nov 24 03:47:11 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/file.h>		/* for R_OK */
#ifndef R_OK
#include <unistd.h>
#endif
#include <fcntl.h>		/* for O_RDWR */
#include <netdb.h>		/* for hostent */
#include <setjmp.h>		/* for jmp_buf */
#include <string.h>		/* for strchr */
#include <pwd.h>		/* for getpwnam */
#include "wfs.h"
#include "util.h"
#include "global.h"
#include <arpa/nameser.h>
#include <resolv.h>

/* configurable parameters */
char		*cs_domain;		/* my domain name */
char		*cs_topdir;		/* top directory of entire system */
unsigned long	cs_minfree;		/* minimum free space to be left */

/* network related information */
char		cs_hostname[MAXHOSTNAMELEN];	/* my hostname */
struct in_addr	cs_ipaddr;			/* my IP address */
int		cs_nfsport;			/* my NFS port */

/* for messages */
char		*cs_progname;			/* my program name */
int		cs_world = 0xC0FFEE;		/* my world id */
int		cs_pid;				/* my process id */
int		cs_uid;				/* my user id */
int		cs_gid;				/* my group id */
char		*cs_conf = "/etc/csd.conf";	/* daemon configuration file */

/* internal state */
serv_state	cs_state;			/* my state */
jmp_buf		select_intr;
int		select_intr_valid;
time_t		now;
int		foreground = 1;
int		immediate_abort;

static void
sigterm(sig)
int sig;
{
	switch (sig) {
	case SIGINT:
		immediate_abort = 15;
		break;
	case SIGTERM:
		immediate_abort = -1;
		break;
	}
	syslog(LOG_WARNING, "going down on signal %d", sig);
	if (select_intr_valid)
		longjmp(select_intr, sig);
}

static void
sighup(sig)
int sig;
{
}

static void
csd_getnames()
{
	/* get identifiers: [gup]id, IP address, and hostname */
	struct sockaddr_in sin;
	struct passwd *pw;

	/*
	 * get uid, gid, pid
	 */
	pw = getpwnam("wwfs");
	if (!pw) {
		syslog(LOG_CRIT, "Password entry for user \"wwfs\" not set");
		csd_abort(1);
	}
	cs_uid = pw->pw_uid;
	cs_gid = pw->pw_gid;
	cs_pid = getpid();

	/*
	 * get IP address
	 */
	get_myaddress(&sin);		/* rpc(3N) */
	cs_ipaddr.s_addr = sin.sin_addr.s_addr;

	/*
	 * get hostname
	 */
	if (gethostname(cs_hostname, sizeof(cs_hostname)) < 0) {
		syslog(LOG_CRIT, "gethostname: %m");
		csd_abort(1);
	}
	if (!*cs_hostname) {
		syslog(LOG_CRIT, "host name is not set");
		csd_abort(1);
	}
}

static void
csd_getdomain()
{
	/*
	 * get network name
	 */
	char *domdot;

	if (domdot = strchr(cs_hostname, '.')) {
		/*
		 * Hostname already contains domainname.
		 * Split out hostname and domainname components
		 */
		*domdot++ = '\0';
		cs_domain = domdot;
	} else if (! cs_domain) {
		/* LIBS=-lresolv and proper DNS configuration required */
		struct hostent *hp;
		hp = gethostbyname(cs_hostname);
		assert(hp != NULL);
		strcpy(cs_hostname, hp->h_name);
		domdot = strchr(cs_hostname, '.');
		if (domdot) {
			*domdot++ = '\0';
			cs_domain = domdot;
		} else {
			int i;
			for (i = 0; hp->h_aliases[i]; ++i) {
				strcpy(cs_hostname, hp->h_aliases[i]);
				domdot = strchr(cs_hostname, '.');
				if (domdot) {
					*domdot++ = '\0';
					cs_domain = domdot;
					break;
				}
			}
		}
		assert(cs_domain != NULL);
	}
}

main(argc, argv)
int argc;
char **argv;
{
	int error;

	/*
	 * initial status
	 */
	cs_state = Start;
	openlog("csd", LOG_PID, LOG_DAEMON);

	/*
	 * sanity check
	 */
	assert(sizeof(nfscookie) >= sizeof (unsigned int));
	assert(sizeof(int) >= 4);
	if (geteuid() != 0) {
		syslog(LOG_WARNING, "Must be root to run WWFS");
		csd_abort(1);
	}

	/*
	 * get program name
	 */
	if (argv[0]) {
		cs_progname = strrchr(argv[0], '/');
		if (cs_progname && cs_progname[1])
			cs_progname++;
		else
			cs_progname = argv[0];
	}
	if (!cs_progname)
		cs_progname = "csd";

	/*
	 * get initial time
	 */
	realtime();

	/*
	 * hook signal
	 */
	/* Trap interrupts for shutdowns. */
	(void) signal(SIGINT, sigterm);

	/* Hangups tell us to reload the cache */
	(void) signal(SIGHUP, sighup);

	/* Trap Terminate so that we can shutdown gracefully (some chance) */
	(void) signal(SIGTERM, sigterm);

	/* ICMP is much informative than SIGPIPE */
	(void) signal(SIGPIPE, SIG_IGN);

	/*
	 * get identifiers
	 */
	csd_getnames();
	if (parse_conf(cs_conf, csd_tailor) < 0) {
		dlog("error reading config file %s", cs_conf);
		csd_abort(1);		/* Fatal error in configuration file */
	}
	csd_getdomain();

	/*
	 * get arguments
	 */
	/* XXX */
	/* get_args(argc, argv); */;

	if (argc == 1	/* XXX */
	    && getppid() != 1) {	/* I'm not started by init. */
		int fd;

		/* portions of this code came from
		 * "UNIX Network Programming", W. Richard Stevens
		 */
		if (background() > 0)
			exit(0);	/* parent process */
		
#ifdef TIOCNOTTY
		/* BSDism here. */
		setpgrp(0, getpid());
		if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
			ioctl(fd, TIOCNOTTY, 0);
			close(fd);
		}
#else
		/* SysV anyone? */
		setpgrp();
#endif
		closelog();
		for (fd = 0; fd < NOFILE; ++fd)
			close(fd);
		errno = 0;
		openlog("csd", LOG_PID, LOG_DAEMON);	/* itch. */
	}

	/*
	 * initialize other module as uid=root
	 */
	csd_start_1();

	/*
	 * change user and group id
	 */
	setuid(cs_uid);
	setgid(cs_gid);

	/*
	 * initialize other module as uid=wwfs
	 */
	csd_start_2();

	/*
	 * other unix environments
	 */
	umask(022);
	chdir(cs_topdir);

	/*
	 * DNS should not freeze csd
	 */
	_res.retrans = 1;
	_res.retry   = 5;

	/*
	 * RPC binding
	 */
	error = csd_rpc();

	/*
	 * start the server
	 */
	if (!error && csd_run() != Done) {
		syslog(LOG_CRIT, "csd_run failed");
		cs_state = Done;
	}

	/*
	 * going down
	 */
	csd_abort(error);

	abort();
}

/* initialize other modules as root */
void
csd_start_1()
{
	dlog_start();
	icmp_start();
}

/* initialize other modules as wwfs */
void
csd_start_2()
{
	getdate();
	alloc_start();
	root_start();
	bfs_start();
	dir_start();
	conn_start();
	thrd_start();
	uip_start();
	talk_start();
#ifdef TRACE
	trace_start();
#endif
}

void
csd_stop()
{
	thrd_stop();
	conn_stop();
	dlog_stop();
}

static int
csd_set_domain(name)
char *name;
{
	if (! strchr(name, '.')) {
		syslog(LOG_ERR, "%s: DOMAIN: invalid domain name", cs_conf);
		return -1;
	}
	cs_domain = name;
	return 0;
}

static int
csd_set_topdir(topdir)
char *topdir;
{
	if (validate_dir(topdir) < 0) {
		syslog(LOG_ERR, "%s: WWFSDIR (%s): %m", cs_conf, topdir);
		return -1;
	}
	/*
	 * now that topdir exists, check if it is writable by "wwfs"
	 */
	if (access(topdir, R_OK|W_OK|X_OK) < 0) {
		syslog(LOG_ERR, "%s: WWFSDIR (%s): permission denied",
		       cs_conf, topdir);
		return -1;
	}
	cs_topdir = topdir;
	return 0;
}

static int
csd_set_minfree(capacity)
char *capacity;
{
	switch (capacity[strlen(capacity)-1]) {
	case 'G':
	case 'g':
		cs_minfree = atoi(capacity) * 1024 * 1024 * 1024;
		break;
	case 'M':
	case 'm':
		cs_minfree = atoi(capacity) * 1024 * 1024;
		break;
	case 'K':
	case 'k':
		cs_minfree = atoi(capacity) * 1024;
		break;
	default:
		syslog(LOG_ERR, "%s: MINFREE: Ambiguous storage unit", cs_conf);
		return -1;
	}
	return 0;
}

static wf_tailor cs_tailor[] = {
	{ "DOMAIN", csd_set_domain },
	{ "MINFREE", csd_set_minfree },
	{ "WWFSDIR", csd_set_topdir },
	{ NULL, (int (*)())0 }
};

int
csd_tailor(name, value)
char *name;
char *value;
{
	return parse_tailor(cs_tailor, name, value);
}

