/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* root of the world view */
static char *AtFSid = "$Header: root.c[109.3] Thu Dec 16 20:39:22 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/types.h>
#ifdef _POSIX_SOURCE
#include <dirent.h>
#define	DIRENT	struct dirent
#else
#include <sys/dir.h>
#define	DIRENT	struct direct
#endif

#include <sys/stat.h>
#include <fcntl.h>		/* for O_WRONLY etc. */
#include <string.h>		/* for strchr */
#include "wfs.h"
#include "util.h"
#include "global.h"

void root_getattr_miss(), root_lookup_miss(), root_read_miss(), root_getuda();

wf_proto proto_root = {
	"root", cmfs_getattr, root_getattr_miss,
	cmfs_lookup, root_lookup_miss, cmfs_readlink,
	cmfs_read, root_read_miss, cmfs_readdir, err_readdir_miss,
	root_getuda, err_keepalive, err_close, err_shutdown, 0
};

/*
 * ID mappings are somewhat tricky:
 * parent_dir == child_dir == WF_ROOT_DIR_ID
 */

wf_vol	*root_vol;
wf_dir	*root_dir;
static wf_file	*root_file;
static fattr	*root_fattr;

/* general functions, not bound to NFS, goes here */

static void
root_load(volp, dotdotp)
wf_vol *volp;
wf_dir *dotdotp;
{
	DIR *dp;
	struct direct *p;
	struct stat statbuf;
	long dirid;
	wf_dir *dirp;
	wf_file *filep, *f;
	char path[MAXPATHLEN], path2[MAXPATHLEN];

	vol_destroyid(volp);		/* clean up bitmap first */
	sprintf(path, "%s/vol/%s", cs_topdir, volp->name);
	dp = opendir(path);
	while (p = readdir(dp)) {
		if (p->d_name[0] == '.')
			continue;	/* ignore "." and ".." */
		sprintf(path2, "%s/vol/%s/%s",
			cs_topdir, volp->name, p->d_name);
		stat(path2, &statbuf);
		if (statbuf.st_mode & S_IFDIR) {
			dirid = vol_allocid(volp) + WF_GLUE_BASE;
			dirp = dir_new(volp, dirid, "/", p->d_name);
			f = dir_adddir(dirp, ".", volp->id, dirid);
			attr_touch(&f->fattr);
			f = dir_adddir(dirp, "..", volp->id, dotdotp->id);
			attr_touch(&f->fattr);
			dirp->flag |= WF_DIR_WIREDDOWN;
			filep = file_newvol(p->d_name);
			FSID(filep) = volp->id;
			FID(filep) = filep->fattr.fileid = dirid;
			attr_touch(&filep->fattr);
			dir_addfile(dotdotp, filep);

			/*
			 * insert index file
			 */
			filep = file_new("INDEX", 0, NFREG, 1, 0644, 0, 0,
					 0, now, 0, volp->id);
			filep->flag |= WF_FILE_MUTABLE;
			dir_addfile(dirp, filep);
		}
	}
	closedir(dp);
}

void
root_start()
{
	wf_vol *volp;
	wf_dir *dirp;
	wf_file *filep;
	char path[MAXPATHLEN];

	/*
	 * create root volume
	 */
	root_vol = volp = vol_alloc();
	volp->id = WF_ROOT_VOL_ID;
	volp->name = ".";			/* name2ih need it */
	volp->dir = "";				/* name2ih need it */
	/*
	 * Bypass file caching mechanism
	 * for files directly placed in the root volume, such as INDEX.
	 */
	volp->flag |= WF_VOL_DIRECTFILE;
	sprintf(path, "%s/vol", cs_topdir);
	volp->topdir = strdup(path);
	/*
	 * create root directory
	 */
	root_dir = dirp = dir_new(volp, WF_ROOT_DIR_ID, 0, 0);
	dir_adddir(dirp, ".", WF_ROOT_VOL_ID, WF_ROOT_DIR_ID);
	dir_adddir(dirp, "..", WF_ROOT_VOL_ID, WF_ROOT_DIR_ID);

	/*
	 * create root node
	 */
	root_file = file_findid(dirp, WF_ROOT_VOL_ID, WF_ROOT_DIR_ID);
	root_fattr = &root_file->fattr;
	attr_touch(root_fattr);
	dirp->flag |= WF_DIR_WIREDDOWN;

	/*
	 * create subdirectories
	 */
	root_load(root_vol, root_dir);

	/*
	 * insert index file
	 */
	filep = file_new("INDEX", 0, NFREG, 1, 0644, 0, 0,
			 0, now, 0, volp->id);
	filep->flag |= WF_FILE_MUTABLE;
	dir_addfile(dirp, filep);
}

wf_file *
root_attach(dirp, name)
wf_dir *dirp;
char *name;
{
	/*
	 * Attach a volume to the specified directory.
	 */
	wf_vol *volp;
	wf_file *filep;
	int ret;
	char *p, *volume;

	p = strrchr(name, '/');
	volume = p ? p+1 : name;
	volp = vol_findname(dirp->name, volume);
	if (volp == (wf_vol *)0)
		return (wf_file *)0;

	filep = file_newvol(volume);
	FSID(filep) = volp->id;
	FID(filep) = filep->fattr.fileid = volp->id + WF_VOL_BASE;
	dir_addfile(dirp, filep);
	return filep;
}

int
root_deletevol(name)
char *name;
{
	/*
	 * Delete a volume from the root namespace.
	 */
	wf_file *filep;
	wf_ih ih;
	int ret;

	ret = root_name2ih(name, &ih);
	if (ret != WF_REP_OK)
		return ret;
	if (ih.filep) {
		dir_deletefile(ih.filep);
		file_uncache(ih.dirp);
		file_free(ih.filep);
	}
	vol_free(ih.child_volp);
	return WF_REP_OK;
}

int
root_fhn2ih(pname, ihp)
char *pname;
wf_ih *ihp;	/* result */
{
	/* convert "+fh/name" into internal handle */
	char *p = pname, *q;
	long pvid, pfid, cvid, cfid;

	/*
	 * see bfs.c:fh_read() for ASCII representation of fh
	 */
	pvid = strtol(++p, &q, 16);
	if (p == q)
		return WF_ERR_STALE;	/* stale ascii file handle */

	pfid = strtol(p = ++q, &q, 16);
	if (p == q)
		return WF_ERR_STALE;	/* stale ascii file handle */

	cvid = strtol(p = ++q, &q, 16);
	ihp->parent_volp = vol_findid(cvid);
	if (! ihp->parent_volp || p == q)
		return WF_ERR_STALE;	/* stale ascii file handle */

	cfid = strtol(p = ++q, &q, 16);
	ihp->dirp = dir_findcache(ihp->parent_volp, cfid);
	if (p == q)
		return WF_ERR_STALE;	/* stale ascii file handle */
	if (! ihp->dirp || (ihp->dirp->flag & WF_DIR_OBSOLETE)) {
		/* Directory not cached. This is likely to happen
		 * when only /a is cached upon request "/a/b/c".
		 * Forcibly ignore the following path components...
		 */
		ihp->child_volp = ihp->parent_volp;
		ihp->dirp = dir_findcache(ihp->parent_volp, pfid);
		if (! ihp->dirp)
			return WF_ERR_STALE;
		ihp->filep = file_findid(ihp->dirp, ihp->child_volp->id, cfid);
		if (! ihp->filep)
			return WF_ERR_STALE;
		return WF_REP_WAIT;
	}

	if (*q != '/')
		return WF_ERR_NOTDIR;
	if (*++q == '\0') q = ".";	/* "fh/" --> "fh/." */
	ihp->filep = file_findname(ihp->dirp, q);
	if (! ihp->filep) {
		if (ihp->dirp->flag & WF_DIR_WIREDDOWN) {
			/* XXX assuming WF_DIR_WIREDDOWN = root/glue */
			/*
			 * looking at invisible volumes...
			 * let's make it visible.
			 */
			ihp->filep = root_attach(ihp->dirp, q);
		}
		if (! ihp->filep) {
			return WF_ERR_NOENT;
		}
	}
	ihp->child_volp = vol_findid(FSID(ihp->filep));
	if (! ihp->child_volp) {
		return WF_ERR_NOENT;
	}
	return WF_REP_OK;
}

int
root_name2ih(pname, ihp)
char *pname;
wf_ih *ihp;	/* result */
{
	char *p, *q = pname;
	wf_vol *volp = root_vol;
	wf_dir *dirp = root_dir, *dirp2 = root_dir;
	wf_file *filep = root_file;
	long prev_vol = volp->id;
	int ret = WF_REP_OK;

	ihp->parent_volp = volp;
	while (q) {
		p = q;
		/* end of path? */
		while (*p == '/') ++p;
		if (*p == '\0') break;
		dirp = dirp2;
		ihp->parent_volp = volp;
		/* lookup a path component */
		q = strchr(p, '/');
		if (q) *q = '\0';
		filep = file_findname(dirp, p);
		if (q) *q = '/';
		if (!filep) {
			if (dirp->flag & WF_DIR_WIREDDOWN) {
				/* XXX assuming WF_DIR_WIREDDOWN = root/glue */
				/*
				 * looking at invisible volumes...
				 * let's make it visible.
				 */
				if (q) *q = '\0';
				filep = root_attach(dirp, p);
				if (q) *q = '/';
			}
			if (!filep) {
				ret = WF_ERR_NOENT;
				goto fin;
			}
		}
		if (filep->fattr.type == NFREG) break;
		if (prev_vol != FSID(filep)) {
			volp = vol_findid(FSID(filep));
			if (!volp) {
				dlog("root_name2ih: illegal fsid");
				ret = WF_ERR_STALE;
				goto fin;
			}
			prev_vol = volp->id;
		}
		dirp2 = dir_findcache(volp, filep->id);
		if (!dirp2 || (dirp2->flag & WF_DIR_OBSOLETE)) {
			/* Directory not cached. This is likely to happen
			 * when only /a is cached upon request "/a/b/c".
			 * Forcibly ignore the following path components...
			 */
			ret = WF_REP_WAIT;
			goto fin;
		}
	}
 fin:
	ihp->child_volp = volp;
	ihp->dirp = dirp;
	ihp->filep = filep;
	return ret;
}

void
ih2fh(ihp, fhp)
wf_ih *ihp;
wf_fh *fhp;	/* result */
{
	fhp->world_id = cs_world;
	fhp->parent_vol = ihp->parent_volp->id;
	fhp->dir_id = ihp->dirp->id;
	fhp->child_vol = ihp->child_volp->id;
	fhp->file_id = ihp->filep->id;
}

int
root_name2fh(pname, fhp)
char *pname;
wf_fh *fhp;	/* result */
{
	int ret;
	wf_ih ih;

	ret = root_name2ih(pname, &ih);
	if (ret == WF_REP_OK) {
		ih2fh(&ih, fhp);
	}
	return ret;
}

void
root_inspect(reqp, arg)
wf_req *reqp;
char *arg;
{
	int ret;
	wf_fh fh;
	int id = reqp->id;

	ret = root_name2fh(arg, &fh);
	if (ret != WF_REP_OK) {
		req_send(reqp, "513.%03d root_name2fh: error code %d.\n",
			 id, ret);
		return;
	}
	req_send(reqp, "220-%03d fhd %s\n", id, arg);
	req_send(reqp, "213-%03d world_id = %lx\n", id, fh.world_id);
	req_send(reqp, "213-%03d parent_vol = %lx\n", id, fh.parent_vol);
	req_send(reqp, "213-%03d dir_id = %lx\n", id, fh.dir_id);
	req_send(reqp, "213-%03d child_vol = %lx\n", id, fh.child_vol);
	req_send(reqp, "213.%03d file_id = %lx\n", id, fh.file_id);
}

/* NFS-delegate functions */

void
root_getattr_miss(c)
wf_thrd *c;	/* volp, dirp, filep */
{
	fattr *attrp = &c->filep->fattr;
	struct stat statbuf;
	char path[MAXPATHLEN];

	file_getpath(path, c->child_volp, c->dirp, c->filep);
	if (stat(path, &statbuf) == 0) {
		attrp->nlink = statbuf.st_nlink;
		attrp->size = statbuf.st_size;
		/* sgi doesn't have st_blocks, so... */
		attrp->blocks = (statbuf.st_size + 1024) >> 10;
		attrp->atime.seconds = statbuf.st_atime;
		attrp->mtime.seconds = statbuf.st_mtime;
		attrp->ctime.seconds = statbuf.st_ctime;
	} else {
		/* File does not exist; fake it. */
		attrp->nlink = 1;
		attrp->size = 0;
		/* sgi doesn't have st_blocks, so... */
		attrp->blocks = 1;
		attrp->atime.seconds = attrp->mtime.seconds = 
			attrp->ctime.seconds = now;
	}
}

void
root_read_miss(c)
wf_thrd *c;
/* in: rqstp, volp, offset, count, totalcount, dirp, filep */
{
	FILE *fp;
	char path[MAXPATHLEN];

	if (strcmp(c->filep->name, "INDEX") == 0) {
		/* INDEX does not exist; generate it on the fly... */
		file_getpath(path, c->child_volp, c->dirp, c->filep);
		fp = fopen(path, "w");
		if (! fp) {
			dlog("Could not open %s for writing: errno=%d",
			     path, errno);
			(*c->reply)(c, WF_ERR_IO);
			return;
		}
		fprintf(fp,
			"Sorry, but INDEX cannot be generated automatically.\n\
Please e-mail your WWFS Administrator <wwfs@%s.%s>\n\
and ask him/her for help.\n\
\n\
Thank you.\n\
		\"csd\" running on %s.%s\n\
		WorldWide File System\n\
		A research product of Osaka University, Japan\n",
			cs_hostname, cs_domain,
			cs_hostname, cs_domain);
		fclose(fp);
		/* try to read me again... */
		(*c->reply)(c, WF_ERR_STALE);
	} else {
		/* strange.. real file should not exist except INDEX */
		(*c->reply)(c, WF_ERR_IO);
	}
}

void
root_lookup_miss(c)
wf_thrd *c;	/* rqstp, volp, fhp, fname, dotdotp */
{
	wf_vol		*subvolp;
	wf_fh		*resfhp;

	c->parent_volp = c->child_volp;
	c->child_volp = vol_findname(c->dotdotp->name, c->fname);
	if (c->child_volp == NULL) {
		(*c->reply)(c, WF_ERR_NOENT);
		return;
	}

	/* update root fattr */
	++root_fattr->nlink;
	attr_touch(root_fattr);

	/* filehandle response */
	resfhp = c->resfhp = CLONE(c->fhp);
	c->resfhp->world_id = cs_world;
	c->resfhp->parent_vol = c->parent_volp->id;
	c->resfhp->parent_dir = c->fhp->child_dir;
	c->resfhp->child_vol = c->child_volp->id;
	c->resfhp->child_dir = c->child_volp->id + WF_VOL_BASE;
	/* make it visible */
	c->filep = file_newvol(c->fname);
	FSID(c->filep) = c->child_volp->id;
	FID(c->filep) = c->filep->fattr.fileid = c->resfhp->file_id;
	dir_addfile(c->dotdotp, c->filep);
	(*c->reply)(c, WF_REP_OK);
	FREE(resfhp);
}

void
root_getuda()
{
	/* XXX not implemented yet */
}
