#!/usr/local/bin/perl
#
# gget               --- recursively get files starting at a given URL
#
# Given a starting URL, gget will recursively retrieve gopher files.
# It will also generate "gget.log" -- a table of titles and hosts.
# Tries to find as many new hosts as possible, and strictly
# limits the number of pages it will request from any one server.
#
# gget will stop when $maxtotal pages are retrieved (1000!),
# or when the all pages were retrieved,
# or when SIGINT is received (^C).
#
# NB: to get all pages recursively, try:
#       gget -ls <home-page>
#
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# derived from "explore" written by: Oscar Nierstrasz oscar@cui.unige.ch
# This file is part of WWFS.
#
#v = '(v1.0)'; # August 30, 1993
#v = '(v1.1)'; # August 31 -- added triggering of xmosaic
#v = '(v1.2)'; # Sept 1 -- added various options; SIGINT handling
#v = '(v1.3)'; # Oct 21 -- fixed counting of hosts; added -d
#v = '(v1.4)'; # Oct 23 -- fixed sigint to allow <CR> to continue
#                #       -- fixed printing of $hostsig
$v = '(v1.0)'; # Nov 14, 1993 -- gget initial revision

require '/etc/csd.pl';
unshift(@INC, "$WWFSDIR/lib");
require 'url.pl';
require 'dirutil.pl';

$usg = 'Usage: gget [<options>] <gopher-url>
	<gopher-url>      -- URL to start with (no default)
	<output file>   -- default is gget.log
	-m <maxpages>   -- max pages to get per site (default 5)
	-t <maxtotal>   -- max total pages to get (default 100)
	-h <maxhosts>   -- max hosts to explore (default unlimited)
	-s <savedir>	-- directory to save pages to (default current)
	-ls             -- list all pages at starting site (use with care!)
';
$maxpages = 5;          # max pages to retrieve per site
$maxtotal = 100;        # max pages to retrieve in total
$maxhosts = undef;      # max hosts to visit

$hosts = 1;             # hosts visited (always at least 1)

chop($date = `date +%d.%m.%y`);
$sig = "This page was generated by gget $v on $date.\n";

# A good default starting point:
# $whatsnew = "http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/whats-new.html";
# $start = "http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/StartingPoints/NetworkStartingPoints.html";

# default log file:
$deflog = "gget.log";

# default save directory:
$savedir = ".";

while ($#ARGV >= $[) {
	$arg = shift @ARGV;

	if ($arg eq "-m") {
		$arg = shift @ARGV;
		if ($arg =~ /^\d+$/){
			print STDERR "maxpages = $arg (was $maxpages)\n";
			$maxpages = $arg;
			next;
		}
		else { die "Bad arg for -m\n$usg"; }
	}

	if ($arg eq "-t") {
		$arg = shift @ARGV;
		if ($arg =~ /^\d+$/){
			print STDERR "maxtotal = $arg (was $maxtotal)\n";
			$maxtotal = $arg;
			next;
		}
		else { die "Bad arg for -t\n$usg"; }
	}

	if ($arg eq "-h") {
		$arg = shift @ARGV;
		if ($arg =~ /^\d+$/){
			print STDERR "maxhosts = $arg (was $maxhosts)\n";
			$maxhosts = $arg;
			next;
		}
		else { die "Bad arg for -h\n$usg"; }
	}

	if ($arg eq "-s") {
		$arg = shift @ARGV;
		if (-d $arg) {
			print STDERR "directory = $arg (was $savedir)\n";
			$savedir = $arg;
			next;
		}
		else { die "Bad arg for -s\n$usg"; }
	}

	if ($arg eq "-ls") {
		$maxhosts = 1;
		$maxtotal = $maxpages = 1000;
		next;
	}

	if ($arg eq "-h") { die "$usg"; }
	if ($arg =~ /^-/) { die "Invalid flag\n$usg"; }
	if ($arg =~ /^gopher:/) {
		if ($start) { die "Please give only one URL\n$usg"; }
		$start = $arg;
	}
	else {
		if ($log) { die "Please give only one output file\n$usg"; }
		$log = $arg;
	}
}

unless ($start) { die "$usg"; }
unless ($log) { $log = $deflog; }
(open(STDOUT,">$log")) || die "Can't create $log\n";
$| = 1;
print STDERR "Writing output to $log\n";
&explore($start);

# Explore the gopherspace, starting at $url.
# Maintains a list @tocheck of URLs to try.
sub explore {
	local($url) = @_;
	if ($maxhosts == 1) {
		print "Log of gopher transfer from $url\n$sig\n";
	}
	else {
		print "Log of gopher transfer starting at $url\n";
		if ($maxhosts) {
			print "Maximum hosts to visit = $maxhosts.\n";
		}
		print "Max pages per site = $maxpages.\n",
			"Max total pages = $maxtotal.\n$sig\n";
	}

	$SIG{'INT'} = 'sigint'; # Stop when SIGINT is received
	push(@tocheck,$url);    # Initialize
	$seen{$url} = 1;        # Remember that we've seen it
	($thistype,$thishost, $thisport, $thispath, $request) =
		&url'parse(undef,undef,undef,undef,$url);
	$seenhost{$thishost} = 1;

	while ($#tocheck >= $[) {
		$url = shift @tocheck;
		# Remember current host, port and path:
		($thistype,$thishost, $thisport, $thispath, $request) =
			&url'parse(undef,undef,undef,undef,$url);
		print STDERR "Requesting $url\n";
		unless ($page = &url'get($url)) {
			print "Can't get $url\n";
			next;
		} ;

		# Extract the title and detect errors
		if ($page =~ /^0Server Error:/) {
			print "Invalid page: $url\n";
			next;
		}

		# Save URL and gopher output
		&save_url($savedir . $request, $url);
		if ($thispath =~ m:^/1/: || $thispath =~ m:^/$:) {
			&savefile("/tmp/gget$$", $page);
		} else {
			&savefile($savedir . $request, $page);
			next;
		}

		# Attempt to convert the gopher listing to an HTML
		open(G, "/tmp/gget$$");
		$html = "Select one of:<P>\n<UL>";
		while (<G>) {
			chop; chop;        # strip CRLF
			($host, $port, $path, $type, $title)
			   = &parse_gopher($_);
			$href = "gopher://$host:$port/$path";
			$html .= "<LI> <A HREF=\"$href\">$title</A>\n";

			# Skip this host if invalid:
			unless ($host) { next; }
			# Skip if seen already:
			if ($seen{$href}) { next; }
			$seen{$href} = 1;
			# Don't ask too many pages from a given host:
			unless (++$count{$host} <= $maxpages) { next; }
			if ($seenhost{$host}) {
				# Seen this host, so add to end of queue:
				print STDERR "Pushing $href\n";
				push(@tocheck,$href);
			}
			else {
				if ($maxhosts) {
					if (++$hosts > $maxhosts) { next; }
				}
				# New host, so add to front:
				print STDERR "Queueing $href\n";
				unshift(@tocheck,$href);
				$seenhost{$host} = 1;
			}
		}
		close(G);
		$html .= "</UL>\n";
		&savefile($savedir . $request . "/", $html);

		if (!($maxhosts == 1)) { $hostsig = " ($thishost)"; }
		# This page is ok, so log it:
		print "$title$hostsig\n";
		# from the last request
		if (++$entries >= $maxtotal) { last; }
	}

	print "\nSearch completed.\n";
	close(STDOUT);
	print STDERR "Result of exploration in $log\n";
}

sub parse_gopher {
	local ($entry) = @_;
	local ($type, $title, $path, $host, $port);

	($title, $path, $host, $port) = split("\t", $entry);
	($type, $title) = unpack("a1 a*", $title);
#	print "$entry => $host, $port, $path, $type, $title\n";
	return ($host, $port, $path, $type, $title);
}

sub save_url {
	# keep url-to-file mapping information
	local ($path, $url) = @_;
	local ($dir, $file);

	($dir, $file) = &basename($path);
	($url, $file) = &basename($url);
	if (! -e "$dir/.url") {
		open(URL, ">$dir/.url");
		print URL "$url\n";
		close(URL);
	}
}

sub savefile {
	local ($path, $page) = @_;
	local ($dir, $file);

	($dir, $file) = &basename($path);
	# make directories if necessary, and write the page
	if (! -d $dir) {
		if (&mkdirhier($dir)) {
			print "mkdir $dir\n";
		} else {
			print "Error in creating $dir: $!\n";
		}
	}
	if ($path =~ m:/$:) {
		# gopherd may have mapped directory to a page,
		# e.g, "/" -> "/foo/Welcome.html".
		# as a workaround, we save it under some name
		# unlikely to conflict.
		$path .= "urlget_dir.html";
	}
	if (open(FILE, ">$path")) {
		print FILE $page;
		close(FILE);
		print "got $path: ";
	} else {
		print "Error in creating $path: $!: ";
	}
}

sub sigint {
	local ($/) = "\n";
	local ($res);
	print STDERR "Enter <CR> to continue, \"q\" to quit\n";
	chop ($res = <STDIN>);
	return if ($res eq "");
	if ($res ne "q") {
	    print STDERR "Invalid response -- continuing\n";
	    return;
	}
	print STDERR "Quitting\n";
	# should dump @tocheck in a file?
	print "\n\nInterrupted!\n";
	close(STDOUT);
	print STDERR "Result of exploration in $log\n";
	exit(0);
}

# return a list of all the hrefs in a page
sub hrefs {
	local($page) = @_;
	$page =~ s/^[^<]+</</;
	$page =~ s/>[^<]*</></g;
	$page =~ s/<a[^>]*href\s*=\s*"([^"]+)"[^>]*>/$1\n/gi;
	$page =~ s/<[^>]*>//g;
	$page =~ s/\n+/\n/g;
	split(/\n/,$page);
}

# Local Variables:
# mode: cperl
# cperl-indent-level: 8
# cperl-continued-statement-offset: 8
# End:
