#! /usr/local/bin/perl
#
#	csd-nanny --- csd OKIAGARI KOBOUSHI Project   Version 1.4
#
# Author: DEMIZU Noritoshi <demizu@nff.ncl.omron.co.jp>
#
# People trying to port csd to new platforms will find this script
# extremely useful --- it was originally written to port csd on
# Luna-II/Mach at OMRON Corporation.
#
# This perl script does the following things:
# 1. If csd ran already, just quit.
# 2. Start csd.
# 3. Wait for the csd to core dump.
# 4. Inspects core file and then takes snapshot of the debug session.
# 5. Send resulting report mail to $csd_admmin.
# 6. goto 2
#
# Debug session consists of: 1) stack trace using gdb, 2) last some lines
# of wwfs.log.  Site administrators might want to insert dlog() around
# the source code so that internal status can be obtained from wwfs.log.

require '/etc/csd.pl';


# site configuration part
$csd_path	= "$WWFSDIR/bin/csd";		# path of csd
$csd_log	= "$WWFSDIR/log/wwfs.log";	# log of csd
$csd_core	= "$WWFSDIR/core";		# core of csd
$csd_uip_port	= 8002;				# csd UIP port number
$csd_tmp	= "/tmp/csd-nanny-work$$";	# working file to report
$csd_admin	= "wwfs";			# e-mail address of admin

$log_lines	= 200;				# reported lines from wwfs.log
$log_expired	= 28;				# older files will be expired

$dbg_path	= "/usr/local/bin/gdb";		# debugger name
$dbg_options	= "-q";				# options for the debugger
$dbg_command	= "where\nquit\n";		# commands to the debugger
$dbg_prompt	= "\\(gdb\\)";			# prompt of the debugger


# check whether csd runs already or not.
if (&check_csd) {
    print STDERR "csd is already running.\n";
    exit 1;
}

# become daemon
$nanny_pid = fork;
if ($nanny_pid < 0) {
    die "can't fork $!, stopped";
} elsif ($nanny_pid > 0) {
    exit 0;
}
close(STDIN); close(STDOUT); close(STDERR);


# infinite loop to keep csd running.
for (;;) {
    &launch;	# launch csd
    &notify;	# notify this accident to csd admin.
    &clean;	# remove old logs and cores
    sleep 3;	# notify this accident to wwfs user :-)
}

exit 1;		# will never reach.


# launch csd and probe it periodically.
sub launch {
    local($pid);

    # check whether csd runs already or not.
    if (&check_csd) {
	print STDERR "csd is already running.\n";
	exit 1;
    }

    # launch csd and wait for the csd to core dump.
    $pid = fork;
    die "can't fork $!, stopped" if ($pid < 0);
    if ($pid == 0) {
	# child process
	exec $csd_path, "foreground";
	die "can't exec $csd_path: $!, stopped";	# will never reach
    } else {
	# parent process
	waitpid($pid, 0);				# wait for child
	#&probe;
    }
}


# notify this accident to csd administrator
# by sending report on csd stack trace and tailing $log_lines lines of wwfs.log
sub notify {
    local($year, $mon, $day, $hour, $min, $sec)
					= (localtime(time))[5,4,3,2,1,0];
    local($date) = sprintf("%02d%02d%02d-%02d%02d%02d",
			$year, $mon+1, $day, $hour, $min, $sec);

    &dbg_info($date);
    system("tail -$log_lines $csd_log >> $csd_tmp");
    system("/usr/ucb/mail -s csd-down-info.$date $csd_admin < $csd_tmp");
    unlink($csd_tmp);

    rename($csd_log, "$csd_log.$date")   if (-f $csd_log);
    rename($csd_core,"$csd_core.$date")  if (-f $csd_core);
}



# save debugging information into a file "$csd_tmp",
# if core and debugger were available.
sub dbg_info {
    local($id) = @_;
    local(@cmds, $cmd);

    return if (!-r $csd_core || !-x $dbg_path);
    open(DEBUG, "| $dbg_path $dbg_options $csd_path $csd_core > $csd_tmp-d")
								|| return;
    print DEBUG $dbg_command;
    close(DEBUG);

    if (!defined($dbg_prompt) || $dbg_prompt eq "") {
	rename("$csd_tmp-d", $csd_tmp);
	return;
    }

    @cmds = split(/\n/, $dbg_command);
    open(TMPD, "$csd_tmp-d") || return;
    open(TMP, ">$csd_tmp") || return;
    print TMP "% $dbg_path $dbg_options $csd_path $csd_core.$id\n";
    while (<TMPD>) {
	if (/($dbg_prompt)/i) {
	    # replace prompt with prompt+command
	    $cmd = shift(@cmds);
	    s/($dbg_prompt)\s+/$1 $cmd\n/i;
	}
	print TMP;
    }
    print TMP "\n\n\n";
    print TMP "---- $csd_log.$id\n";
    close(TMPD);
    close(TMP);
    unlink("$csd_tmp-d");
}


# checks whether csd is already running or not.
# return code means:
#	1: csd is running already
#	0: csd isn't running yet.
sub check_csd {
    require "sys/socket.ph";
    socket(SOCK, &AF_INET, &SOCK_STREAM, 0) || die "socket: $!";
    $dstaddr = pack('S n a4 x8', &AF_INET,
		$csd_uip_port, (gethostbyname("localhost"))[4]);
    $result = connect(SOCK, $dstaddr);
    print SOCK "QUIT\n" if ($result);
    close(SOCK);
    return $result;
}


sub clean {
    while (<$csd_log.*>) {
	unlink if (-M $_ > $log_expired);
    }
    while (<$csd_core.*>) {
	unlink if (-M $_ > $log_expired);
    }
}
